unit facedetect;

interface

uses Windows, SysUtils, Classes, Graphics, opencv;

type
  TCamView = class
  private
    frontalface_cascade: PCvHaarClassifierCascade;
    capture: PCvCapture;
    FImage: PIplImage; // last captured frame //
    storage: PCvMemStorage; // mem storage for obj detection //
    FWidth: integer;
    FHeight: integer;
    FObjectFound: boolean;
    FObjectFoundTimes: integer;
    FObjectX: integer;
    FObjectY: integer;
    FObjectRect: CvRect;
    // for face recognition //
    FFacesCount: integer;
    FFacesCollection: array [0..100] of PIplImage;
  public
    property Width: integer read FWidth;
    property Height: integer read FHeight;
    property ObjectFound: boolean read FObjectFound;
    property ObjectX: integer read FObjectX;
    property ObjectY: integer read FObjectY;
    property ObjectRect: CvRect read FObjectRect;

    constructor Create;
    destructor Destroy; override;

    procedure Open(cam_index: integer);
    procedure Close;
    procedure GetFrame; // captures the frame into FImage //
    procedure RunFace; // searches for faces and gets the biggest one (frontal face preferrable) //
    procedure ToBitmap(var bmp: _SimpleBitmap); // converts IplImage to bitmap //
    procedure SaveFace(filename: string);
    // returns face filename if found //
    // first call GetFrame, then RunFace, finally call RecognizeFace //
    function RecognizeFace: string;
  end;

const
  FRONTAL_FACE = 'haarcascades\haarcascade_frontalface_alt.xml';

implementation
//------------------------------------------------------------------------------
constructor TCamView.Create;
var
  i: integer;
  img_load: PIplImage;
begin
  FObjectFoundTimes:= 0;
  frontalface_cascade:= cvLoad(pchar(FRONTAL_FACE), nil, nil, nil);
  if not assigned(frontalface_cascade) then Exception.Create('ERROR: Could not load frontalface classifier cascade');
  storage:= cvCreateMemStorage(4194304);

  // load faces collection //
  FFacesCount:= 21;
  try
    for i:= 0 to FFacesCount - 1 do
    begin
      img_load:= cvLoadImage(pchar('faces\' + inttostr(i) + '.png'), CV_LOAD_IMAGE_COLOR);
      FFacesCollection[i]:= cvCreateImage(cvSize_(img_load^.width, img_load^.height), 8, 1);
      cvCvtColor(img_load, FFacesCollection[i], CV_BGR2GRAY);
    end;
  except
    on e: Exception do
    begin
      messagebox(0, pchar('load faces'#13#10#13#10 + e.message), nil, 0);
      exit;
    end;
  end;
end;
//------------------------------------------------------------------------------
destructor TCamView.Destroy;
begin
  close;
  if assigned(storage) then cvReleaseMemStorage(@storage);
end;
//------------------------------------------------------------------------------
procedure TCamView.Open(cam_index: integer);
begin
  capture:= cvCreateCameraCapture(cam_index);
  if not assigned(capture) then
  begin
    Exception.Create('ERROR: Could not create cam capture');
    exit;
  end;

  GetFrame;
end;
//------------------------------------------------------------------------------
procedure TCamView.Close;
var
  i: integer;
begin
  if assigned(capture) then cvReleaseCapture(@capture);
  FImage:= nil;
  for i:= 0 to FFacesCount - 1 do
  begin
    try cvReleaseImage(@FFacesCollection[i]);
    except end;
  end;
end;
//------------------------------------------------------------------------------
procedure TCamView.GetFrame;
begin
  FImage:= nil;
  if assigned(capture) then
  begin
    FImage:= cvQueryFrame(capture);
    if not assigned(FImage) then
    begin
      Exception.Create('ERROR: Could not capture the frame');
      exit;
    end;
    FWidth:= FImage^.width;
    FHeight:= FImage^.height;
  end;
end;
//------------------------------------------------------------------------------
procedure TCamView.RunFace;
var
  gray: PIplImage;
  seq: PCvSeq;
  r: PCvRect;
  square_coeff: extended;
begin
  FObjectFound:= false;

  if assigned(FImage) then
  begin
    // search biggest face //
    gray:= cvCreateImage(cvSize_(FImage^.width, FImage^.height), 8, 1);
    cvCvtColor(FImage, gray, CV_BGR2GRAY);
    cvEqualizeHist(gray, gray);
    cvClearMemStorage(storage);
    seq:= cvHaarDetectObjects(gray, frontalface_cascade, storage, 1.1, 2, CV_HAAR_FIND_BIGGEST_OBJECT + CV_HAAR_DO_CANNY_PRUNING, cvSize_(20, 20));
    r:= PCvRect(cvGetSeqElem(seq, 0, nil));

    if assigned(r) then
    begin
      // filter off accidential matches //
      try
        square_coeff:= r^.width * r^.height / FObjectRect.width / FObjectRect.width;
        if square_coeff < 1 then square_coeff:= 1 / square_coeff;
        if (square_coeff <= 3) and
            (abs(r^.x + r^.width div 2 - FObjectRect.x - FObjectRect.width div 2) / FObjectRect.width < 0.8) and
            (abs(r^.y + r^.height div 2 - FObjectRect.y - FObjectRect.height div 2) / FObjectRect.height < 0.8)
        then
          inc(FObjectFoundTimes)
        else
          FObjectFoundTimes:= 0;
      except
      end;
      if FObjectFoundTimes > 5 then FObjectFound:= true;

      // store object data //
      FObjectX:= r^.x + r^.width div 2;
      FObjectY:= r^.y + r^.height div 2;
      FObjectRect:= CvRect_(r^.x, r^.y, r^.width, r^.height);
    end;

    if assigned(gray) then cvReleaseImage(@gray);
  end;
end;
//------------------------------------------------------------------------------
procedure TCamView.ToBitmap(var bmp: _SimpleBitmap);
begin
  if assigned(FImage) then IplImage2SimpleBitmap(FImage, bmp);
end;
//------------------------------------------------------------------------------
procedure TCamView.SaveFace(filename: string);
var
  img: PIplImage;
begin
  if FObjectFound then
  begin
    if 1.2 > FObjectRect.height / FObjectRect.width then FObjectRect.height:= round(1.2 * FObjectRect.width)
    else if 1.2 < FObjectRect.height / FObjectRect.width then FObjectRect.width:= round(1.2 * FObjectRect.height);
    img:= cvCreateImage(CvSize_(100, 120), 8, 3);
    cvSetImageROI(FImage, CvRect_(FObjectX - FObjectRect.width div 2, FObjectY - FObjectRect.height div 2, FObjectRect.width, FObjectRect.height));
    cvResize(FImage, img, CV_INTER_LINEAR);
    cvResetImageROI(FImage);
    cvSaveImage(pchar(filename), img);
    cvReleaseImage(@img);
  end;
end;
//------------------------------------------------------------------------------
// returns face filename if found //
function TCamView.RecognizeFace: string;
var
  i, iTrain, iNearest, w, h: integer;
  img_load, test_img, mean_img: PIplImage;
  tc: CvTermCriteria;
  EigenVals: array [0..100] of single;
  projectedTestFace: array [0..100] of single;
  coeffs: array [0..100] of array [0..100] of single;
  eig_img: array [0..100] of PIplImage;
  d_i: single;
  distSq, leastDistSq: double;
begin
  result:= '';

  try
    w:= FObjectRect.width;
    h:= FObjectRect.height;
    if 1.2 > h / w then h:= round(1.2 * w) else if 1.2 < h / w then w:= round(1.2 * h);
    cvSetImageROI(FImage, CvRect_(FObjectX - w div 2, FObjectY - h div 2, w, h));
    img_load:= cvCreateImage(cvsize_(100, 120), 8, 3);
    cvResize(FImage, img_load, CV_INTER_LINEAR);
    cvResetImageROI(FImage);
    test_img:= cvCreateImage(cvsize_(100, 120), 8, 1);
    cvCvtColor(img_load, test_img, CV_BGR2GRAY);
  except
  end;

  try
    mean_img:= cvCreateImage(cvsize_(100, 120), IPL_DEPTH_32F, 1); //  
    for i:= 0 to FFacesCount - 1 do eig_img[i]:= cvCreateImage(cvsize_(100, 120), IPL_DEPTH_32F, 1);
    tc.type_:= CV_TERMCRIT_NUMB + CV_TERMCRIT_EPS;
    tc.max_iter:= 100;
    tc.epsilon:= 0.001;
    cvCalcEigenObjects(FFacesCount, @FFacesCollection, @eig_img, CV_EIGOBJ_NO_CALLBACK, 0, nil, @tc, mean_img, @EigenVals);
  except
  end;

  //      
  try for i:= 0 to FFacesCount - 1 do cvEigenDecomposite(FFacesCollection[i], FFacesCount, @eig_img, CV_EIGOBJ_NO_CALLBACK, nil, mean_img, @coeffs[i]);
  except end;

  //   
  try cvEigenDecomposite(test_img, FFacesCount, @eig_img, CV_EIGOBJ_NO_CALLBACK, nil, mean_img, @projectedTestFace);
  except end;

  try
    leastDistSq:= $fffffff;
    iNearest:= -1;
    for iTrain:= 0 to FFacesCount - 1 do
    begin
      distSq:= 0;
      for i:= 0 to FFacesCount - 1 do
      begin
        d_i:= projectedTestFace[i] - coeffs[iTrain][i];
        distSq:= distSq + d_i * d_i / EigenVals[i];
      end;
      if distSq < leastDistSq then
      begin
        leastDistSq:= distSq;
        iNearest:= iTrain;
      end;
    end;
    if leastDistSq <= 1{threshold} then result:= inttostr(iNearest) + '.png';
  except
  end;

  //  
  for i:= 0 to FFacesCount - 1 do
    try cvReleaseImage(@eig_img[i]);
    except end;

  try
    cvReleaseImage(@img_load);
    cvReleaseImage(@mean_img);
    cvReleaseImage(@test_img);
  except end;
end;
//------------------------------------------------------------------------------
end.
